# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

""" Define some exception classes to aid in testing, i.e to make sure that
    workflows are being rejected for appropriate reasons
"""


class CyclicWorkflowException(Exception):
    pass


class DisconnectedWorkflowException(Exception):
    pass


class DuplicateWorkflowNames(Exception):
    pass


class MissingSubWorkflowReference(Exception):
    pass


class NoPrimaryWorkflows(Exception):
    pass


class MultiplePrimaryWorkflows(Exception):
    pass


class InvalidWorkflowReference(Exception):
    pass


class InvalidNodeReference(Exception):
    pass


class InvalidFlowControlNode(Exception):
    pass


class MissingPreprocessorException(Exception):
    pass


class DuplicateRegistryConfigName(Exception):
    pass


class InvalidConfig(Exception):
    pass
