# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import itertools
import six


def merge_dicts(dicts):
    result = {}

    for _dict in dicts:
        bad_keys = [key for (key, value) in six.iteritems(_dict) if
                    result.get(key, value) != value]
        if bad_keys:
            raise Exception('ERROR: plugins define incompatible values for keys {}'.format(
                bad_keys))

        result.update(_dict)

    return result


def merge_lists(lists):
    return list(itertools.chain.from_iterable(lists))
